use serde::Deserialize;

use crate::image::{Image, Pixel};
use std::collections::{HashMap, HashSet};

#[derive(Deserialize, Debug)]
struct ColorEntry {
    color: Pixel,
    symbol: String,
}

#[derive(Deserialize, Debug)]
pub struct Mapping(HashMap<Pixel, String>);

#[derive(Debug)]
pub struct IncompleteMappingError {
    path: std::path::PathBuf,
    missing_colors: Vec<Pixel>,
}

impl std::fmt::Display for IncompleteMappingError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        writeln!(
            f,
            "`{:?}` is missing entries for the following pixel colors:",
            self.path
        )?;
        for color in self.missing_colors.iter() {
            writeln!(f, "  - {:?}", color)?;
        }
        Ok(())
    }
}

impl std::error::Error for IncompleteMappingError {}

impl Mapping {
    pub fn load(
        path: impl AsRef<std::path::Path>,
        img: &Image,
    ) -> Result<Mapping, Box<dyn std::error::Error>> {
        let path = path.as_ref().to_path_buf();
        let data: Vec<ColorEntry> = serde_yaml::from_reader(std::fs::File::open(&path)?)?;
        // do validation to make sure all pixel colors are handled
        let color_map = data
            .into_iter()
            .map(|entry| (entry.color, entry.symbol))
            .collect::<HashMap<Pixel, String>>();
        let all_image_colors = img
            .iter()
            .map(|(_, color)| color)
            .collect::<HashSet<Pixel>>();
        let missing_colors = all_image_colors
            .into_iter()
            .filter(|c| !color_map.contains_key(&c))
            .collect::<Vec<Pixel>>();
        if !missing_colors.is_empty() {
            return Err(Box::new(IncompleteMappingError {
                path,
                missing_colors,
            }));
        }
        Ok(Mapping(color_map))
    }

    pub fn lookup(&self, color: Pixel) -> &str {
        &self.0[&color]
    }
}

#[derive(Deserialize)]
pub struct Config {
    pub grid_every: u32,
    pub line_weight: f64,
    pub major_line_weight: f64,
    pub grid_size: f64,
    pub font: String,
}

impl Config {
    pub fn scale(&self, n: u32) -> f64 {
        n as f64 * self.grid_size
    }
}

impl Default for Config {
    fn default() -> Config {
        Config {
            grid_every: 10,
            line_weight: 1.0,
            major_line_weight: 3.0,
            grid_size: 24.0,
            font: "Fira Sans 12".to_string(),
        }
    }
}
